@@ -44,6 +44,13 @@
 #define DEBUG 0
 #endif
 
+#ifndef DEBUG_POLYGONS
+#define DEBUG_POLYGONS 0
+#endif
+
+//#define OLD_ATTAINED(X) do { X; } while(0);
+#define OLD_ATTAINED(X) while(0) { X; };
+
 static int compare_x_asc (const void *p1, const void *p2)
 {
     objective_t x1 = **(objective_t **)p1;
@@ -60,28 +67,62 @@ static int compare_y_desc (const void *p1, const void *p2)
     return (y1 > y2) ? -1 : ((y1 < y2) ? 1 : 0);
 }
 
+static inline void
+point2d_printf(FILE *stream, const objective_t x, const objective_t y)
+{
+    fprintf(stream, point_printf_format "\t" point_printf_format, x, y);
+}
+
+static inline void
+point_printf(FILE *stream, const objective_t *p, int nobj)
+{
+    point2d_printf(stream, p[0], p[1]);
+    for (int k = 2; k < nobj; k++)
+        fprintf (stream, "\t" point_printf_format, p[k]);
+}
+
 eaf_t * eaf_create (int nobj, int nruns, int npoints)
 {
     eaf_t *eaf;
-    EAF_MALLOC (eaf, 1, sizeof(eaf_t));
+    EAF_MALLOC (eaf, 1, eaf_t);
     eaf->nobj = nobj;
     eaf->nruns = nruns;
     eaf->size = 0;
+    eaf->nreallocs = 0;
     /* Maximum is npoints, but normally it will be smaller, so at most
        log2(2 * nruns) realloc will occur.  */
     eaf->maxsize = 256 + npoints / (2 * nruns); 
-
-    EAF_MALLOC (eaf->data, nobj * eaf->maxsize, sizeof(objective_t));
-    EAF_MALLOC (eaf->attained, nruns * eaf->maxsize, sizeof(bool));
+    /* fprintf(stderr,"maxsize %ld = %d npoints, %d nruns\n", */
+    /*         eaf->maxsize, npoints, nruns); */
+    EAF_MALLOC (eaf->data, nobj * eaf->maxsize, objective_t);
+    eaf->bit_attained = malloc (bit_array_bytesize(nruns) * eaf->maxsize);
+    eaf->attained = NULL;
+    OLD_ATTAINED(eaf->attained = malloc(sizeof(bool) * nruns * eaf->maxsize));
     return eaf;
 }
 
 void eaf_delete (eaf_t * eaf)
 {
     free (eaf->data);
-    free (eaf->attained);
+    OLD_ATTAINED(free (eaf->attained));
+    free (eaf->bit_attained);
     free (eaf);
 }
+void eaf_realloc(eaf_t * eaf, size_t nobj)
+{
+    const int nruns = eaf->nruns;
+    eaf->data = realloc (eaf->data,
+                         sizeof(objective_t) * nobj * eaf->maxsize);
+    eaf_assert(eaf->data);
+    OLD_ATTAINED(
+        eaf->attained = realloc (eaf->attained, 
+                                 sizeof(bool) * nruns * eaf->maxsize);
+        eaf_assert(eaf->attained);
+        );
+    eaf->bit_attained = realloc (eaf->bit_attained, 
+                                 bit_array_bytesize(nruns) * eaf->maxsize);
+    eaf_assert(eaf->bit_attained);
+}
 
 objective_t *
 eaf_store_point_help (eaf_t * eaf, int nobj,
@@ -91,21 +132,39 @@ eaf_store_point_help (eaf_t * eaf, int nobj,
 
     if (eaf->size == eaf->maxsize) {
         eaf_assert (eaf->size < INT_MAX / 2);
-        eaf->maxsize = eaf->maxsize * 2;
-        eaf->attained = realloc (eaf->attained, 
-                                 sizeof(bool) * nruns * eaf->maxsize);
-        eaf_assert(eaf->attained);
-        eaf->data = realloc (eaf->data,
-                             sizeof(objective_t) * nobj * eaf->maxsize);
-        eaf_assert(eaf->data);
+        //size_t old_maxsize = eaf->maxsize;
+        eaf->maxsize = (size_t) (eaf->maxsize * (1.0 + 1.0 / pow(2, eaf->nreallocs / 4.0)));
+        eaf->maxsize += 100; // At least we increase it by 100 points
+        /* fprintf(stderr,"maxsize (%d): %ld -> %ld\n", eaf->nreallocs, */
+        /*         old_maxsize, eaf->maxsize); */
+        eaf->nreallocs++;
+        // FIXME: We could save memory by only storing eaf->attained per point if requested.
+        eaf_realloc(eaf, nobj);
     }
-    /* We convert from int to bool to hopefully save space. */
+    // FIXME: provide a bit_array function to do this.
     for (int k = 0; k < nruns; k++) {
-        eaf->attained[(nruns * eaf->size) + k] = (bool) save_attained[k];
+        bit_array_set(bit_array_offset(eaf->bit_attained, eaf->size, nruns), k, (bool) save_attained[k]);
+        OLD_ATTAINED(eaf->attained[nruns * eaf->size + k] = (bool) save_attained[k]);
     }
+    OLD_ATTAINED(
+    bitset_check(bit_array_offset(eaf->bit_attained, eaf->size, eaf->nruns),
+                 eaf->attained + eaf->size * eaf->nruns, eaf->nruns));
+ 
     return eaf->data + nobj * eaf->size;
 }
 
+static void
+eaf_adjust_memory (eaf_t * eaf, int nobj)
+{
+    if (eaf->size < eaf->maxsize) {
+        //fprintf(stderr,"reduce size: %ld -> %ld\n", eaf->maxsize, eaf->size);
+        eaf->maxsize = eaf->size;
+        eaf_realloc(eaf, nobj);
+    }
+    OLD_ATTAINED(
+        bit_array_check(eaf->bit_attained, eaf->attained, eaf->size, eaf->nruns));
+}
+
 static void
 eaf_store_point_2d (eaf_t * eaf, objective_t x, objective_t y, 
                     const int *save_attained)
@@ -120,18 +179,14 @@ eaf_store_point_2d (eaf_t * eaf, objective_t x, objective_t y,
 static void
 eaf_print_line (FILE *coord_file, FILE *indic_file, FILE *diff_file, 
                 const objective_t *x, int nobj,
-                const bool *attained, int nruns)
+                const bit_array *attained, int nruns)
 {
     int count1 = 0;
     int count2 = 0;
     int k;
 
     if (coord_file) {
-        fprintf (coord_file, point_printf_format "\t" point_printf_format,
-                 (double)x[0], (double)x[1]);
-        for (k = 2; k < nobj; k++)
-            fprintf (coord_file, "\t" point_printf_format, (double)x[k]);
-
+        point_printf(coord_file, x, nobj);
         fprintf (coord_file, 
                  (coord_file == indic_file) || (coord_file == diff_file)
                  ? "\t" : "\n");
@@ -139,13 +194,13 @@ eaf_print_line (FILE *coord_file, FILE *indic_file, FILE *diff_file,
 
     if (indic_file) {
         fprintf (indic_file, "%d", 
-                 attained[0] ? (count1++,1) : 0);
+                 bit_array_get(attained, 0) ? (count1++,1) : 0);
         for (k = 1; k < nruns/2; k++) 
             fprintf (indic_file, "\t%d", 
-                     attained[k] ? (count1++,1) : 0);
+                     bit_array_get(attained, k) ? (count1++,1) : 0);
         for (k = nruns/2; k < nruns; k++)
             fprintf (indic_file, "\t%d", 
-                     attained[k] ? (count2++,1) : 0);
+                     bit_array_get(attained, k) ? (count2++,1) : 0);
 
         fprintf (indic_file, (indic_file == diff_file) ? "\t" : "\n");
     } else if (diff_file) {
@@ -160,14 +215,30 @@ eaf_print_line (FILE *coord_file, FILE *indic_file, FILE *diff_file,
 void
 eaf_print_attsurf (eaf_t * eaf, FILE *coord_file,  FILE *indic_file, FILE *diff_file)
 {
-    int i;
-    for (i = 0; i < eaf->size; i++) {
+    OLD_ATTAINED(
+    bit_array_check(eaf->bit_attained,
+                    eaf->attained, eaf->size, eaf->nruns));
+      
+    for (size_t i = 0; i < eaf->size; i++) {
         const objective_t *p = eaf->data + i * eaf->nobj;
+        /* bit_array_fprintf(stderr, eaf->bit_attained, eaf->nruns * eaf->size); */
+        OLD_ATTAINED(bitset_check(bit_array_offset(eaf->bit_attained, i, eaf->nruns),
+                                  eaf->attained + i * eaf->nruns, eaf->nruns));
         eaf_print_line (coord_file, indic_file, diff_file,
-                        p, eaf->nobj, eaf->attained + i * eaf->nruns, eaf->nruns);
+                        p, eaf->nobj,
+                        bit_array_offset(eaf->bit_attained, i, eaf->nruns),
+                        eaf->nruns);
     }
 }
 
+static inline void
+fprint_set2d (FILE *stream, const objective_t * const *data, int ntotal)
+{
+    for (int k = 0; k < ntotal; k++)
+        fprintf (stream, "%6d: " point_printf_format " " point_printf_format "\n", k,
+                 data[k][0], data[k][1]);
+}
+
 /* 
    eaf2d: compute attainment surfaces from points in objective space,
           using dimension sweeping.
@@ -197,7 +268,7 @@ eaf2d (const objective_t *data, const int *cumsize, int nruns,
     
     const int ntotal = cumsize[nruns - 1]; /* total number of points in data */
     int *runtab;	
-    int *attained, nattained, *save_attained;
+    int *attained, *save_attained;
     int k, j, l;
 
     /* Access to the data is made via two arrays of pointers: ix, iy
@@ -211,7 +282,7 @@ eaf2d (const objective_t *data, const int *cumsize, int nruns,
 
 #if DEBUG > 1
     fprintf (stderr, "Original data:\n");
-    fprint_set (stderr, datax, ntotal);
+    fprint_set2d (stderr, datax, ntotal);
 #endif
 
     qsort (datax, ntotal, sizeof(*datax), &compare_x_asc);
@@ -219,9 +290,9 @@ eaf2d (const objective_t *data, const int *cumsize, int nruns,
 
 #if DEBUG > 1
     fprintf (stderr, "Sorted data (x):\n");
-    fprint_set (stderr, datax, ntotal);
+    fprint_set2d (stderr, datax, ntotal);
     fprintf (stderr, "Sorted data (y):\n");
-    fprint_set (stderr, datay, ntotal);
+    fprint_set2d (stderr, datay, ntotal);
 #endif
 
     /* Setup a lookup table to go from a point to the approximation
@@ -253,13 +324,12 @@ eaf2d (const objective_t *data, const int *cumsize, int nruns,
         int level = attlevel[l];
         int x = 0;
         int y = 0;
-        int run;
-
-        nattained = 0;
+        
+        int nattained = 0;
         for (k = 0; k < nruns; k++) attained[k] = 0;
 
         /* Start at upper-left corner */
-        run = runtab[(datax[x] - data) / nobj];
+        int run = runtab[(datax[x] - data) / nobj];
         attained[run]++;
         nattained++;
 
@@ -314,6 +384,7 @@ eaf2d (const objective_t *data, const int *cumsize, int nruns,
                                 save_attained);
 
         } while (x < ntotal - 1 && y < ntotal);
+        eaf_adjust_memory(eaf[l], nobj);            
     }
     free(save_attained);
     free(attained);
@@ -325,14 +396,19 @@ eaf2d (const objective_t *data, const int *cumsize, int nruns,
 }
 
 
-#define PRINT_POINT(X,Y) fprintf(stderr, point_printf_format "\t" point_printf_format "\n", X, Y)
-#undef  PRINT_POINT
-#define PRINT_POINT(X,Y) (void)0
+#if DEBUG_POLYGONS > 0
+#define PRINT_POINT(X,Y,C) do { \
+    fprintf(stdout, "PRINT_POINT:"); point2d_printf(stdout, X, Y);             \
+    if (C != INT_MIN) fprintf(stdout, "\t%d", C);                              \
+    fprintf(stdout, "\n"); } while(0)
+#else 
+#define PRINT_POINT(X,Y,C) (void)0
+#endif
 
 static int
 eaf_max_size(eaf_t * const * eaf, int nlevels)
 {
-    int max_size = 0;
+    size_t max_size = 0;
     for (int a = 0; a < nlevels; a++) {
         if (max_size < eaf[a]->size)
             max_size = eaf[a]->size;
@@ -343,19 +419,144 @@ eaf_max_size(eaf_t * const * eaf, int nlevels)
 static int
 eaf_diff_color(const eaf_t * eaf, size_t k, int nruns)
 {
-    const bool *attained = eaf->attained + k * nruns;
+    const bit_array *bit_attained = bit_array_offset(eaf->bit_attained, k, nruns);
+    OLD_ATTAINED(
+        const bool *attained = eaf->attained + k * nruns;
+        bitset_check(bit_attained, attained, nruns););
     int count_left, count_right;
-    attained_left_right (attained, nruns/2, nruns, &count_left, &count_right);
+    attained_left_right (bit_attained, nruns/2, nruns, &count_left, &count_right);
     return count_left - count_right;
 }
 
-void
+static void
 init_colors(int * color, const eaf_t * eaf, size_t eaf_size, int nruns)
 {
     for (size_t k = 0; k < eaf_size; k++) {
         color[k] = eaf_diff_color(eaf, k, nruns);
     }
 }
+
+static const objective_t *
+next_polygon(const objective_t *src, int nobj, const objective_t * end)
+{
+    while (src < end && *src != objective_MIN)
+        src += nobj;
+    src += nobj;
+    return src;
+}
+
+static void
+min_max_in_objective(const objective_t *v,
+                     int nobj, int k,
+                     objective_t *min_ref, objective_t *max_ref)
+{
+    eaf_assert(k < nobj);
+    eaf_assert(v[k] != objective_MIN); // Empty polygon?
+
+    objective_t min = v[k];
+    objective_t max = v[k];
+    v += nobj;
+    while (v[k] != objective_MIN) {
+        if (min > v[k]) min = v[k];
+        if (max < v[k]) max = v[k];
+        v += nobj;
+    }
+    *min_ref = min;
+    *max_ref = max;
+}
+
+static bool
+polygon_dominates_point(const objective_t *p, const objective_t *x, int nobj)
+{
+    eaf_assert(x[0] != objective_MIN);
+    while (p[0] != objective_MIN) {
+        if (p[0] < x[0] && p[1] < x[1]) {
+            /* point2d_printf(stdout, p[0], p[1]); */
+            /* printf(" dominates "); */
+            /* point2d_printf(stdout, x[0], x[1]); */
+            /* printf("\n"); */
+            return true;
+        }
+        p += nobj;
+    }
+    return false;
+}
+
+static bool
+polygon_dominates_any_point(const objective_t *a, const objective_t *b, int nobj)
+{
+    while (b[0] != objective_MIN) {
+        if (polygon_dominates_point(a, b, nobj)) return true;
+        b += nobj;
+    }
+    return false;
+}
+
+static bool
+polygons_intersect(const objective_t *a, const objective_t *b, int nobj)
+{
+    for (int k = 0; k < nobj; k++) {
+        objective_t min_a, max_a, min_b, max_b;
+        min_max_in_objective(a, nobj, k, &min_a, &max_a);
+        min_max_in_objective(b, nobj, k, &min_b, &max_b);
+        // If we can draw a line completely separating them in one axis, then
+        // they don't intersect.
+        if (max_a <= min_b || max_b <= min_a) return false;
+    }
+    // Two orthogonal polygons intersect if there is a corner of A that is
+    // dominated by a corner of B and there is a corner of B that is dominated
+    // by a corner of A
+    return polygon_dominates_any_point(a, b, nobj) && polygon_dominates_any_point(b, a, nobj);
+}
+
+
+static inline void
+polygon_print(const objective_t *p, int nobj)
+{
+    while (p[0] != objective_MIN) {
+        point_printf(stderr, p, nobj);
+        fprintf(stderr, "\n");
+        p += nobj;
+    }
+    point_printf(stderr, p, nobj);
+    fprintf(stderr, "\n");
+}
+
+__unused static void 
+eaf_check_polygons(eaf_polygon_t *p, int nobj)
+{
+    // This only works for 2 objectives.
+    assert(nobj == 2);
+        
+    // Check #1: Polygons don't intersect
+    // Last point of last polygon
+    const objective_t * end = vector_objective_end(&p->xy);
+    const objective_t * pi = vector_objective_begin(&p->xy);
+    
+    while(pi < end) {
+        const objective_t * pj = next_polygon(pi, nobj, end);
+        const objective_t * next = pj;
+        while(pj < end) {
+            if (polygons_intersect(pi, pj, nobj)) {
+#if DEBUG_POLYGONS > 0
+                fprintf(stderr, "ERROR: Polygons intersect!\n");
+                polygon_print(pi, nobj);
+                polygon_print(pj, nobj);
+#endif
+                eaf_assert(false);
+            }
+            pj = next_polygon(pj, nobj, end);
+        }
+        pi = next;
+    }
+    // Check #2: Every point in the EAF is a corner of a polygon and it has the
+    // same color as the polygon.
+
+    //TODO
+
+    
+}
+
 /* Produce a polygon suitable to be plotted by the polygon function in R.  */
 eaf_polygon_t *
 eaf_compute_polygon (eaf_t **eaf, int nlevels)
@@ -366,17 +567,26 @@ eaf_compute_polygon (eaf_t **eaf, int nlevels)
     do { _poly_size_check--; eaf_assert(_poly_size_check >= 4);                \
         eaf_assert(_poly_size_check % 2 == 0); _poly_size_check = 0; } while(0)
 #define eaf_point(A,K) (eaf[(A)]->data + (K) * nobj)
-#define push_point(X, Y)                                                       \
-    do {  vector_objective_push_back (&polygon->xy, (X));                      \
-        vector_objective_push_back (&polygon->xy, (Y));                        \
-        _poly_size_check++; PRINT_POINT(X,Y);                                  \
+#define push_point_color(X, Y, C)                                              \
+        do {  vector_objective_push_back (&polygon->xy, (X));                  \
+            vector_objective_push_back (&polygon->xy, (Y));                    \
+            _poly_size_check++; PRINT_POINT(X,Y, C);                           \
     } while(0)
-    
-#define polygon_close()                                                        \
-    do {                                                                       \
-        vector_int_push_back (&polygon->col, color_0);                         \
-        push_point(objective_MIN, objective_MIN); POLY_SIZE_CHECK(); } while(0)
 
+#define push_point(X, Y) push_point_color(X,Y, INT_MIN)
+
+#if DEBUG_POLYGONS > 0
+#define EXPENSIVE_CHECK_POLYGONS() eaf_check_polygons(polygon, nobj)
+#else
+#define EXPENSIVE_CHECK_POLYGONS() (void)0
+#endif
+
+#define polygon_close(COLOR) do {                                              \
+                vector_int_push_back (&polygon->col, COLOR);                   \
+                push_point_color(objective_MIN, objective_MIN, COLOR);         \
+                POLY_SIZE_CHECK(); EXPENSIVE_CHECK_POLYGONS();                 \
+            } while(0)
+    
     int _poly_size_check = 0;
     int nruns = eaf[0]->nruns;
     int nobj = eaf[0]->nobj;
@@ -385,9 +595,9 @@ eaf_compute_polygon (eaf_t **eaf, int nlevels)
 
     int max_size = eaf_max_size(eaf, nlevels);
     int *color;
-    EAF_MALLOC (color, sizeof(int), max_size);
+    EAF_MALLOC (color, max_size, int);
     eaf_polygon_t * polygon;
-    EAF_MALLOC(polygon, sizeof(eaf_polygon_t), 1);
+    EAF_MALLOC(polygon, 1, eaf_polygon_t);
     vector_objective_ctor (&polygon->xy, max_size);
     vector_int_ctor (&polygon->col, max_size);
  
@@ -405,11 +615,13 @@ eaf_compute_polygon (eaf_t **eaf, int nlevels)
             const objective_t * pka = NULL;
             const objective_t * pkb = NULL;
             int kb = last_b + 1;
+            // Find a point in a that does not overlap with a point in b
             while (ka < eaf_a_size && kb < eaf_b_size) {
                 pka = eaf_point (a, ka);
                 pkb = eaf_point (b, kb);
                 if (pkb[0] != pka[0])
                     break;
+                // They overlap in x, so we will skip kb, remember it.
                 topleft_y = pkb[1];
                 last_b = kb;
                 if (pkb[1] == pka[1]) {
@@ -459,7 +671,8 @@ eaf_compute_polygon (eaf_t **eaf, int nlevels)
                         break;
                     }
                 }
-            } while (ka < eaf_a_size && color_0 != color[ka]);
+                /* FIXME: When is color_0 != color[ka] if they are in the same eaf level ? */ 
+            } while (ka < eaf_a_size && color_0 == color[ka]);
 
             /* pka is the point before changing color, but ka is the
                position after.  */
@@ -468,8 +681,9 @@ eaf_compute_polygon (eaf_t **eaf, int nlevels)
                    completely dominates eaf_b, so just start by the
                    end. */
                 if (last_b == eaf_b_size - 1) {
-                    /* There is nothing on the other side, just create
-                       two points in the infinity. */
+                    /* The last point we skipped was the last point, so there
+                       is nothing on the other side, just create two points in
+                       the infinity. */
                     push_point (objective_MAX, pka[1]);
                     push_point (objective_MAX, topleft_y);
                     eaf_assert(topleft_y > pka[1]);
@@ -477,19 +691,26 @@ eaf_compute_polygon (eaf_t **eaf, int nlevels)
                     kb = eaf_b_size - 1;
                     pkb = eaf_point(b, kb);
                     eaf_assert (pkb[1] >= pka[1]);
-                    eaf_assert (pkb[0] > pka[0]);
                     if (pkb[1] > pka[1]) {
-                        /* Create two points in the infinity.  */
+                        // Create two points in the infinity to jump from a to b (turn the corner).
+                        /* If pkb is above pkda, then it may happen that pkb[0]
+                           <= pka[0] if pkb is dominated by a previous pka. */
                         push_point (objective_MAX, pka[1]);
                         push_point (objective_MAX, pkb[1]);
                         eaf_assert (pkb[1] <= topleft_y);
+                    } else {
+                        // If they are at the same y-level, then pka must be to
+                        // the left, otherwise we should have found the
+                        // intersection earlier and something is wrong.
+                        eaf_assert(pkb[0] > pka[0]);
                     }
+                    /* Now print in reverse.  */
                     objective_t prev_pkb_x = pkb[0];
                     push_point (pkb[0], pkb[1]);
                     kb--;
                     while (kb > last_b) {
                         pkb = eaf_point(b, kb);
-                        eaf_assert (pkb[0] > pka[0]);
+                        eaf_assert (pkb[1] > pka[1]); // pkb cannot be below pka
                         push_point (prev_pkb_x, pkb[1]);
                         push_point (pkb[0], pkb[1]);
                         prev_pkb_x = pkb[0];
@@ -499,7 +720,7 @@ eaf_compute_polygon (eaf_t **eaf, int nlevels)
                     eaf_assert (topleft_y > pkb[1]);
                 }
                 /* last_b = eaf_b_size - 1; */
-                polygon_close(); /* DONE */
+                polygon_close(color_0); /* DONE */
                 break; /* Really done! */
             } else {
                 if (kb == eaf_b_size) { 
@@ -510,38 +731,44 @@ eaf_compute_polygon (eaf_t **eaf, int nlevels)
                     push_point (objective_MAX, topleft_y);
                     last_b = eaf_b_size - 1;
                 } else {
+                    // This polygon is bounded above by eaf_b from last_b up to kb.
                     eaf_assert (kb < eaf_b_size);
                     pkb = eaf_point (b, kb);
+                    // If we have not finished eaf_b is because pkb is not above pka
                     eaf_assert (pkb[1] <= pka[1]);
                     eaf_assert (pkb[0] != pka[0]);
                     /* If pkb and pka are in the same horizontal, pkb
                        does not affect the next polygon. Otherwise, it
                        does. */
-                    int save_last_b = (pkb[1] == pka[1]) ? kb : kb - 1;;
-                    /* Now print in reverse.  */
+                    int save_last_b = (pkb[1] == pka[1]) ? kb : kb - 1;
+                    /* Now print eaf_b in reverse.  */
                     objective_t prev_pkb_x = pkb[0];
                     push_point (pkb[0], pka[1]);
                     kb--;
                     while (kb > last_b) {
                         pkb = eaf_point (b, kb);
-                        eaf_assert (pkb[0] != pka[0]);
+                        // pkb must be above pka or we would have found the
+                        // intersection earlier.
+                        eaf_assert (pkb[1] > pka[1]); 
                         push_point (prev_pkb_x, pkb[1]);
                         push_point (pkb[0], pkb[1]);
                         prev_pkb_x = pkb[0];
                         kb--;
                     }
-                    eaf_assert (topleft_y > pkb[1]);
                     push_point (pkb[0], topleft_y);
+                    eaf_assert (topleft_y > pkb[1]);
                     last_b = save_last_b;
                 }
-                polygon_close(); /* DONE */
+                polygon_close(color_0); /* DONE */
                 eaf_assert(topleft_y >= pka[1]);
                 topleft_y = pka[1];
             }
         }
     }
-
     free (color);
+#if DEBUG >= 1
+    eaf_check_polygons(polygon, nobj); // This is slow with lots of polygons
+#endif
     return polygon;
 }
 
@@ -561,8 +788,8 @@ eaf_compute_polygon_old (eaf_t **eaf, int nlevels)
 
     eaf_assert(nruns % 2 == 0);
 
-    EAF_MALLOC (color, sizeof(int), max_size);
-    EAF_MALLOC(polygon, sizeof(eaf_polygon_t), 1);
+    EAF_MALLOC (color, max_size, int);
+    EAF_MALLOC(polygon, 1, eaf_polygon_t);
     vector_objective_ctor (&polygon->xy, max_size);
     vector_int_ctor (&polygon->col, max_size);
  
@@ -623,7 +850,7 @@ eaf_compute_polygon_old (eaf_t **eaf, int nlevels)
                     push_point (pkb[0], topleft_y);
                     last_b = eaf_b_size - 1;
                 }
-                polygon_close(); /* DONE */
+                polygon_close(color_0); /* DONE */
             } else {
                 int kb = last_b + 1;
                 /* Different color, go down by the other side until
@@ -633,7 +860,7 @@ eaf_compute_polygon_old (eaf_t **eaf, int nlevels)
                        two points in the infinity. */
                     push_point (objective_MAX, pka[1]);
                     push_point (objective_MAX, topleft_y);
-                    polygon_close(); /* DONE */
+                    polygon_close(color_0); /* DONE */
                 } else {
                     const objective_t * pkb;
                     do {
@@ -662,7 +889,7 @@ eaf_compute_polygon_old (eaf_t **eaf, int nlevels)
                         } 
                         push_point (pkb[0], topleft_y);
                     }
-                    polygon_close(); /* DONE */
+                    polygon_close(color_0); /* DONE */
                     last_b = save_last_b;
                 }
             }
@@ -685,9 +912,9 @@ eaf_print_polygon (FILE *stream, eaf_t **eaf, int nlevels)
     eaf_polygon_t *p = eaf_compute_area (eaf, nlevels);
     
     for(size_t i = 0; i < vector_objective_size(&p->xy); i += 2) {
-        fprintf(stream, point_printf_format "\t" point_printf_format "\n", 
-                vector_objective_at(&p->xy, i),
-                vector_objective_at(&p->xy, i + 1));
+        point2d_printf(stream, vector_objective_at(&p->xy, i),
+                       vector_objective_at(&p->xy, i + 1));
+        fprintf(stream, "\n");
     }
 
     fprintf (stream, "# col =");
@@ -700,16 +927,25 @@ eaf_print_polygon (FILE *stream, eaf_t **eaf, int nlevels)
     free(p);
 }
 
-size_t
-region_push (vector_objective *regions,
-             objective_t lx, objective_t ly,
-             objective_t ux, objective_t uy)
+static size_t
+rectangle_add(eaf_polygon_t * regions,
+              objective_t lx, objective_t ly,
+              objective_t ux, objective_t uy,
+              int color)
 {
-    vector_objective_push_back(regions, lx);
-    vector_objective_push_back(regions, ly);
-    vector_objective_push_back(regions, ux);
-    vector_objective_push_back(regions, uy);
-    return vector_objective_size(regions);
+#if 0
+    printf("rectangle_add: (" point_printf_format ", " point_printf_format ", " point_printf_format ", " point_printf_format ")[%d]\n", lx, ly, ux, uy, color);
+#endif
+    eaf_assert(lx < ux);
+    eaf_assert(ly < uy);
+    vector_objective *rect = &regions->xy;
+    vector_objective_push_back(rect, lx);
+    vector_objective_push_back(rect, ly);
+    vector_objective_push_back(rect, ux);
+    vector_objective_push_back(rect, uy);
+
+    vector_int_push_back(&regions->col, color);
+    return vector_objective_size(rect);
 }
 
 eaf_polygon_t *
@@ -718,20 +954,21 @@ eaf_compute_rectangles (eaf_t **eaf, int nlevels)
 #define eaf_point(A,K) (eaf[(A)]->data + (K) * nobj)
 #if 0
 #define printf_points(ka,kb,pka,pkb)                                           \
-    printf("%4d: pa[%d]=(" point_printf_format ", " point_printf_format "), pb[%d] = (" point_printf_format ", " point_printf_format ")\n", __LINE__, ka, pka[0], pka[1], kb, pkb[0], pkb[1])
+    printf("%4d: pa[%d]=(" point_printf_format ", " point_printf_format "), pb[%d] = (" point_printf_format ", " point_printf_format ")\n", \
+           __LINE__, ka, pka[0], pka[1], kb, pkb[0], pkb[1])
 #else
 #define printf_points(ka,kb,pka,pkb)                                           
 #endif    
     int nruns = eaf[0]->nruns;
-    int nobj = eaf[0]->nobj;
+    const int nobj = eaf[0]->nobj;
 
     eaf_assert(nruns % 2 == 0);
 
     int max_size = eaf_max_size(eaf, nlevels);
     int *color;
-    EAF_MALLOC (color, sizeof(int), max_size);
+    EAF_MALLOC (color, max_size, int);
     eaf_polygon_t * regions;
-    EAF_MALLOC(regions, sizeof(eaf_polygon_t), 1);
+    EAF_MALLOC(regions, 1, eaf_polygon_t);
     vector_objective_ctor (&regions->xy, max_size);
     vector_int_ctor (&regions->col, max_size);
         
@@ -741,76 +978,56 @@ eaf_compute_rectangles (eaf_t **eaf, int nlevels)
         const int eaf_b_size = eaf[b]->size;
         if (eaf_a_size == 0 || eaf_b_size == 0) continue;
 
+        // FIXME: Skip points with color 0?
         init_colors(color, eaf[a], eaf_a_size, nruns);
         objective_t top = objective_MAX;
         int ka = 0, kb = 0;
-        const objective_t * pka = eaf_point (a, ka);
         const objective_t * pkb = eaf_point (b, kb);
+        const objective_t * pka = eaf_point (a, ka);
         printf_points(ka, kb, pka, pkb);
-        /* printf("attained[ka] ="); */
-        /* for (int k = 0; k < nruns; k++)  */
-        /*     printf(" %d", (eaf[a]->attained + ka * nruns)[k]  ? 1 : 0); */
-        /* printf("\n"); */
-        /* printf("attained[kb] ="); */
-        /* for (int k = 0; k < nruns; k++)  */
-        /*     printf(" %d", (eaf[b]->attained + kb * nruns)[k]  ? 1 : 0); */
-        /* printf("\n"); */
-        /* printf("attained[%d][%d,%d] = [%d, %d]\n", ka, a, b, */
-        /*        (eaf[a]->attained + ka * nruns)[a]  ? 1 : 0, */
-        /*        (eaf[a]->attained + ka * nruns)[b]  ? 1 : 0); */
-        /* printf("attained[%d][%d,%d] = [%d, %d]\n", kb, a, b, */
-        /*        (eaf[b]->attained + kb * nruns)[a]  ? 1 : 0, */
-        /*        (eaf[b]->attained + kb * nruns)[b]  ? 1 : 0); */
-               
-        eaf_assert(pka[0] <= pkb[0]);
-        /* It is possible that pka does not dominate pkb if their intersection
-           belongs to the next eaf level. */
-        if (pka[1] >= pkb[1]) goto pka_above_equal_pkb;
-                
-    pka_below_pkb:
-        eaf_assert(pka[1] < pkb[1]);
-        /* If pka[0] >= pkb[0] then no rectangle is created. */
-        if (pka[0] < pkb[0]) {
-            region_push(&regions->xy, pka[0], pkb[1], pkb[0], top);
-            vector_int_push_back(&regions->col, color[ka]);
-        } 
-        top = pkb[1];
-        kb++;
-        if (kb >= eaf_b_size) goto close_eaf;
-        pkb = eaf_point (b, kb);
-        printf_points(ka, kb, pka, pkb);
-        if (pka[1] < pkb[1]) goto pka_below_pkb;
+        while (true) {
 
-    pka_above_equal_pkb:
-        if (pka[0] < pkb[0]) {                    
-            region_push(&regions->xy, pka[0], pka[1], pkb[0], top);
-            vector_int_push_back(&regions->col, color[ka]);
-        } else {
-            // Handle repeated points
-            eaf_assert(pka[0] == pkb[0] && pka[1] == pkb[1]);
+            while (pka[1] < pkb[1]) {
+                if (pka[0] < pkb[0]) // pka strictly dominates pkb
+                    rectangle_add(regions, pka[0], pkb[1], pkb[0], top, color[ka]);
+                top = pkb[1];
+                kb++;
+                if (kb >= eaf_b_size) goto close_eaf;
+                pkb = eaf_point (b, kb);
+                printf_points(ka, kb, pka, pkb);
+            }
+            // pka_above_equal_pkb:
+            if (pka[0] < pkb[0]) { // pka does not strictly dominate pkb
+                rectangle_add(regions, pka[0], pka[1], pkb[0], top, color[ka]);
+            } else {
+                // Skip repeated points
+                eaf_assert(pka[0] == pkb[0] && pka[1] == pkb[1]);
+            }
             top = pka[1];
-            ka++, kb++;
+            ka++;
             if (ka >= eaf_a_size) goto next_eaf;
-            if (kb >= eaf_b_size) goto close_eaf;
-            
             pka = eaf_point (a, ka);
-            pkb = eaf_point (b, kb);
             printf_points(ka, kb, pka, pkb);
-            eaf_assert(pka[0] <= pkb[0]);
-            if (pka[1] >= pkb[1]) goto pka_above_equal_pkb;
-            else goto pka_below_pkb;
-        }
-        top = pka[1];
-        ka++;
-        if (ka >= eaf_a_size) goto next_eaf;
-        pka = eaf_point (a, ka);
-        printf_points(ka, kb, pka, pkb);
-        if (pka[1] >= pkb[1]) goto pka_above_equal_pkb;
-        else goto pka_below_pkb;
 
+            if (pkb[1] == top) { // pkb was not above but equal to previous pka
+                // Move to next pkb
+                kb++;
+                if (kb >= eaf_b_size) goto close_eaf;
+                pkb = eaf_point (b, kb);
+                printf_points(ka, kb, pka, pkb);
+            }
+        }
     close_eaf:
-        region_push(&regions->xy, pka[0], pka[1], objective_MAX, top);
-        vector_int_push_back(&regions->col, color[ka]);
+        // b is finished, add one rectangle for each pka point.
+        while (true) {
+            eaf_assert(pka[1] < pkb[1]);
+            rectangle_add(regions, pka[0], pka[1], objective_MAX, top, color[ka]);
+            top = pka[1];
+            ka++;
+            if (ka >= eaf_a_size) break;
+            pka = eaf_point (a, ka);
+            printf_points(ka, kb, pka, pkb);
+        }
     next_eaf:
         continue;
     }
